/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007, 2008, 2009  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#ifndef PCAP_LIST_OF_TRACES_HXX_
# define PCAP_LIST_OF_TRACES_HXX_

# include "list_of_traces.hh"

namespace pcapxx
{

  template <class Desc, class Bottom>
  list_of_traces<Desc, Bottom>::list_of_traces(const std::string& trace_names)
  {
    typedef std::vector<std::string>			trace_vect_t;
    typedef std::pair<trace_vect_t, std::string>	parse_res_t;

    const parse_res_t	pres = parse(trace_names);
    const trace_vect_t& names = pres.first;

    std::string		file_name;

    if (not pres.second.empty())
      meta_.add("addr", pres.second);
    switch (names.size())
      {
      case 0:
	file_name = "[]";
	break;
      case 1:
	file_name = names.front();
	break;
      default:
	{
	  size_t	len = 0;
	  bool		cont = true;

	  while (cont and len <= names.front().size())
	    {
	      const std::string	ref = names.front().substr(0, ++len);

	      for (size_t i = 1; cont and i < names.size(); ++i)
		if (names[i].substr(0, len) != ref)
		  cont = false;
	    }
	  file_name = '[' + names.front().substr(0, len - 1) + "*]";
	  break;
	}
      }
    meta_.add("file_name", file_name);

    BOOST_FOREACH(const std::string& s, names)
      push_back(Desc (s));

    update_metadata(this);
  }

  template <class Desc, class Bottom>
  list_of_traces<Desc, Bottom>::list_of_traces(const list_of_traces& other):
    super_type (static_cast<const super_type&> (other)),
    meta_ (other.meta_)
  {
    update_metadata(this);
  }

  template<class Desc, class Bottom>
  list_of_traces<Desc, Bottom>&
  list_of_traces<Desc, Bottom>::operator = (const list_of_traces& rhs)
  {
    if (&rhs == this)
      return *this;

    super_type::operator = (rhs);
    meta_ = rhs.meta_;
    update_metadata(this);
  }

  template <class Desc, class Bottom>
  const std::string&
  list_of_traces<Desc, Bottom>::file_name() const
  {
    return meta_["file_name"].get<std::string>();
  }

  template <class Desc, class Bottom>
  const tool::opt::list&
  list_of_traces<Desc, Bottom>::metadata() const
  {
    return meta_;
  }

  template <class Desc, class Bottom>
  void
  list_of_traces<Desc, Bottom>::expect(typename Desc::link_type l) const
  {
    for (size_t i = 0; i < this->size(); ++i)
      this->get(i).expect(l);
  }

  template <class Desc, class Bottom>
  std::pair<std::string, size_t>
  list_of_traces<Desc, Bottom>::parse_word(const std::string&	 trace_names,
					   const size_t		 start,
					   const std::set<char>& stop_seps)
  {
    std::string	w;
    size_t	i;

    for (i = start;
	 i < trace_names.size() and not stop_seps.count(trace_names[i]);
	 ++i)
      {
	if (trace_names[i] == escape_char)
	  ++i;
	if (i < trace_names.size())
	  w.push_back(trace_names[i]);
      }
    return std::make_pair(w, i);
  }

  template <class Desc, class Bottom>
  std::pair<std::vector<std::string>, std::string>
  list_of_traces<Desc, Bottom>::parse(const std::string& traces_names)
  {
    std::vector<std::string>	names;
    std::string			addr;
    std::set<char>		stop_sep_tname;

    stop_sep_tname.insert(trace_name_sep);
    stop_sep_tname.insert(addr_sep);

    size_t	i = 0;

    while (i < traces_names.size() and traces_names[i] != addr_sep)
      {
	std::pair<std::string, size_t>	w = parse_word(traces_names, i,
						       stop_sep_tname);
	names.push_back(w.first);
	i = w.second;
	if (i < traces_names.size() and traces_names[i] == trace_name_sep)
	  ++i;
      }
    if (i < traces_names.size() and traces_names[i] == addr_sep)
      addr = parse_word(traces_names, i + 1, std::set<char> ()).first;
    return std::make_pair(names, addr);
  }

  template <class Desc, class Bottom>
  void
  list_of_traces<Desc, Bottom>::update_metadata(list_of_traces* p)
  {
    for (unsigned i = 0; i < this->size(); ++i)
      this->get(i).metadata().add("list_of_traces", p);
  }

} // End of namespace pcapxx.

#endif // ! PCAP_LIST_OF_TRACES_HXX_
