/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007, 2008, 2009  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#ifndef TOOL_OPTIONS_HH_
# define TOOL_OPTIONS_HH_

# include <stdexcept>
# include <string>
# include <map>
# include <ostream>

# include <boost/any.hpp>


namespace tool
{

  namespace opt
  {

    /*---------------------.
    | Forward declarations |
    `---------------------*/

    namespace internals
    {
      struct pair;
      template <class> struct defaults_list;
    }

    struct name;


    /*-------------.
    | Declarations |
    `-------------*/

    /// Exception raised when converting opt::value to invalid types.
    struct invalid_type: std::logic_error
    {
      invalid_type(const std::string& opt_type_name,
		   const std::string& req_type_name);
    };

    /// Option value for opt::list.
    struct value: public boost::any
    {
      value();

      template <class T>
      value(const T& v);

      template <class T>
      T&
      get();

      template <class T>
      const T&
      get() const;

      value&
      operator [] (const name& n);

      const value&
      operator [] (const name& n) const;

      std::ostream&
      print(std::ostream& v) const;
    };

    std::ostream&
    operator << (std::ostream& o, const value& v);

    /// Convert an option value to a specific type.
    template <class T>	T&		get(value& v);

    /// Convert an option value to a specific type, const version.
    template <class T>	const T&	get(const value& v);

    /// Option name for opt::list.
    struct name: public std::string
    {
      name(const std::string& n);
      name(const char* n);

      internals::pair
      operator -= (const value& v) const;
    };

    /// Option related internals.  Not designed for users.  @see opt::list.
    namespace internals
    {

      /// (name, value) option pair for opt::list's constructor.
      struct pair
      {
	pair(const name& n, const value& v);

	defaults_list<pair>
	operator , (const pair& p) const;

	template <class Map>
	void
	populate(Map& m) const;

      private:
	const name&	opt_name;
	const value&	opt_value;
      };

      /// Static list of opt::internals::pair's.
      template <class NextDefaultsList>
      struct defaults_list
      {
	defaults_list(const pair& p, const NextDefaultsList& next);

	defaults_list< defaults_list<NextDefaultsList> >
	operator , (const pair& p) const;

	template <class Map>
	void
	populate(Map& m) const;

      private:
	const pair&		this_pair;
	const NextDefaultsList&	next_defaults;
      };

    } // End of namespace tool::opt::internals.

    /// Exception raised when accessing a non-available option in opt::list.
    struct invalid_option: public std::invalid_argument
    {
      invalid_option(const name& n);
    };

    /**
     * List of options.
     *
     * A %list of options is a dynamic map that associates values of
     * arbitrary types to std::string keys.  Once a %list of options
     * is constructed and filled with various options, one access
     * a specific item with operator[]():
     *
     * @code
     *   opt::list  l;
     *
     *   l.add("int",   int (42));
     *   l.add("float", float (42));
     *
     *   int   i = l["int"].get<int>();		// Both methods
     *   float f = opt::get<float>(l["float"]);	// are valid.
     * @endcode
     *
     * Instead of using add() it is possible to initialize a %list as
     * follows:
     *
     * @code
     *   opt::list l ((opt::name ("int") -= 42,
     *		       opt::name ("sublist") -= opt::list
     *						((
     *						  opt::name ("bool") -= true
     *						))
     *		     ));
     * @endcode
     *
     * You can access sub-list items as follows:
     *
     * @code
     *   bool b = l["sublist"]["bool"].get<bool>();
     * @endcode
     *
     * Looking for an option that does not exist or trying
     * a conversion to an invalid type with raise an exception.
     */
    struct list: public std::map<name, value>
    {
      typedef std::map<name, value>	super_type;

      /// Default constructor.
      list();

      /// Constructor from expressions of kind "opt::name ("x") -= y".
      list(const internals::pair& defaults);

      /// Constructor from list of expressions of kind "opt::name ("x") -= y".
      template <class D>
      list(const internals::defaults_list<D>& defaults);

      /// Get the option value associated to option name @c n.
      value&
      operator [] (const name& n);

      /// Get the option value associated to option name @c n, const version.
      const value&
      operator [] (const name& n) const;

      /// Add the options from another list directly in this list.
      void
      add(const list& l);

      /// Add an option of name @c n and of value @c v.
      void
      add(const name& n, const value& v);

      /**
       * Add another option list @c l as a new option with name @c n.
       *
       * If an option with name @c n already exists and is already an
       * option list, the options from @c l are *added* to this list,
       * i.e.  the previous options are not deleted in the sub-list.
       */
      list&
      add(const name& n, const list& l);

      /// Fill another option list with the options from the list.
      void
      populate(list& other) const;

      /// Print an option list.
      std::ostream&
      print(std::ostream& o, unsigned depth = 0) const;
    };

    std::ostream&
    operator << (std::ostream& o, const list& l);

  } // End of namespace tool::opt.

} // End of namespace tool.

# include "options.hxx"

#endif // TOOL_OPTIONS_HH_
