/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007, 2008, 2009  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#ifndef WIFI_FRAME_DATA_HH_
# define WIFI_FRAME_DATA_HH_

# include <wipal/wifi/addr.hh>

namespace wifi
{

  namespace frame
  {

    /// IEEE 802.11 <i>%data %frame</i> compounds.
    namespace data
    {

      /// IEEE 802.11 <i>%data %frame</i> subtypes.
      namespace subtype
      {

	/// IEEE 802.11 <i>%data %frame</i> subtypes.
	enum frame_subtype
	  {
	    data		= 0, ///< Data %subtype.
	    data_cf_ack		= 1, ///< Data+CF-Ack %subtype.
	    data_cf_poll	= 2, ///< Data+CF-Poll %subtype.
	    data_cf_ack_cf_poll	= 3, ///< Data+CF-Ack+CF-Poll %subtype.
	    null		= 4, ///< Null-data %subtype.
	    cf_ack		= 5, ///< CF-Ack %subtype.
	    cf_poll		= 6, ///< CF-Poll %subtype.
	    cf_ack_cf_poll	= 7  ///< CF-Ack+CF-Poll %subtype.
	  };

	/// IEEE 802.11 %data %frame subtype names.
	extern const std::string	names[16];

      } // End of namespace wifi::frame::data::subtype.

      /// IEEE 802.11 %data %frame structure (3 addresses).
      struct header_3addr
      {
	uint16_t	frame_ctl;	///< Frame control field.
	uint16_t	duration;	///< Frame duration field.
	addr		addrs[3];	///< Frame addresses.
	uint16_t	seq_ctl;	///< Frame sequence control field.
	uint8_t	body[];			///< Frame body.
      } __attribute__((__packed__));

      /// IEEE 802.11 %data %frame structure (4 addresses).
      struct header_4addr
      {
	uint16_t	frame_ctl;	///< Frame control field.
	uint16_t	duration;	///< Frame duration field.
	addr		addrs[3];	///< Frame's first three addresses.
	uint16_t	seq_ctl;	///< Frame sequence control field.
	addr		addrs_4;	///< Frame's last address (4th).
	uint8_t	body[];			///< Frame body.
      } __attribute__((__packed__));

      /// IEEE 802.11 %data %frame compounds (ToDS = FromDS = 0).
      namespace within_ibss
      {

	/// IEEE 802.11 %data %frame structure (ToDS = FromDS = 0).
	struct header : public header_3addr
	{
	  /// Data %frame addresses index.
	  enum addr_index
	    {
	      da	= 0,	///< Destination address.
	      sa	= 1,	///< Source address.
	      bssid	= 2	///< BSSID.
	    };
	} __attribute__((__packed__));
      } // End of namespace wifi::frame::data::within_ibss.

      /// IEEE 802.11 %data %frame compounds (ToDS = 1, FromDS = 0).
      namespace to_ds
      {

	/// IEEE 802.11 %data %frame structure (ToDS = 1, FromDS = 0).
	struct header : public header_3addr
	{
	  /// Data %frame addresses index.
	  enum addr_index
	    {
	      bssid	= 0,	///< BSSID.
	      sa	= 1,	///< Source address.
	      da	= 2	///< Destination address.
	    };
	} __attribute__((__packed__));
      } // End of namespace wifi::frame::data::to_ds.

      /// IEEE 802.11 %data %frame compounds (ToDS = 0, FromDS = 1).
      namespace from_ds
      {

	/// IEEE 802.11 %data %frame structure (ToDS = 0, FromDS = 1).
	struct header : public header_3addr
	{
	  /// Data %frame addresses index.
	  enum addr_index
	    {
	      da	= 0,	///< Destination address.
	      bssid	= 1,	///< BSSID.
	      sa	= 2	///< Source address.
	    };
	} __attribute__((__packed__));
      } // End of namespace wifi::frame::data::from_ds.

      /// IEEE 802.11 %data %frame compounds (ToDS = FromDS = 1).
      namespace ap_to_ap
      {

	/// IEEE 802.11 %data %frame structure (ToDS = FromDS = 1).
	struct header
	{
	  /// Data %frame addresses index.
	  enum addr_index
	    {
	      ra	= 0,	///< Receiver address.
	      ta	= 1,	///< Transmitter address.
	      da	= 2	///< Destination address.
	    };

	  uint16_t	frame_ctl;	///< Frame control field.
	  uint16_t	duration;	///< Frame duration field.
	  addr		addrs[3];	///< Frame's first three addresses.
	  uint16_t	seq_ctl;	///< Frame sequence control field.
	  addr		addrs_sa;	///< Frame's source address.
	  uint8_t	body[];		///< Frame body.
	} __attribute__((__packed__));

      } // End of namespace wifi::frame::data::ap_to_ap.

    } // End of namespace wifi::frame::data.

  } // End of namespace wifi::frame.

} // End of namespace wifi.

#endif // ! WIFI_FRAME_DATA_HH_
