/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2008, 2009  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#ifndef WIFI_FRAME_FILTER_ANON_POLICIES_HH_
# define WIFI_FRAME_FILTER_ANON_POLICIES_HH_

# include <map>
# include <set>

# include <boost/array.hpp>

# include <wipal/tool/mapping.hh>
# include <wipal/wifi/addr.hh>
# include <wipal/wifi/essid.hh>

namespace wifi
{

  namespace frame
  {

    namespace anon
    {

      /*-----------------------------.
      | Prefix preserving MAC policy |
      `-----------------------------*/

      namespace internals
      {

	struct ppmp_mapping_factory
	{
	  typedef std::map<addr, addr>			mac_map;

	  addr			new_value(const addr&		a,
					  const mac_map&	);
	  void			value_loaded(const addr&	,
					     const addr&	a,
					     const mac_map&	);

	private:
	  std::set<addr>	values_;
	};

      } // End of namespace internals.

      typedef
        tool::mapping<addr, addr, internals::ppmp_mapping_factory>
        prefix_preserving_mac_policy;

      /*-------------------------------.
      | Prefix preserving ESSID policy |
      `-------------------------------*/

      namespace internals
      {

	struct ppep_mapping_factory
	{
	  typedef std::map<essid, essid>	essid_map;

	  ppep_mapping_factory();

	  essid	new_value(const essid&		e,
			  const essid_map&	m);
	  void	value_loaded(const essid&	from,
			     const essid&	to,
			     const essid_map&	);
	private:
	  // Forward declaration.
	  struct node;

	  std::pair<std::string, node*>
	  longest_anonymized_prefix(const essid& orig_essid);


	  /*------.
	  | Nodes |
	  `------*/

	  struct node
	  {
	    node*			child(unsigned char c);
	    std::pair<char, node*>	add_valued_child(unsigned char c);
	    node*			add_valued_child(unsigned char c,
							 unsigned char value);

	    virtual			~node();

	  private:
	    static unsigned char	rand_extended();
	    static unsigned char	rand_alnum();
	    static unsigned char	rand_print();
	    static unsigned char	rand_other();
	    static unsigned char	rand_same_class(unsigned char c);

	  private:
	    boost::array<std::auto_ptr<node>, 256>	children_;
	    std::set<char>				children_values_;
	  };

	  struct valued_node: public node
	  {
	    valued_node(char value);

	    char		value;
	  };

	  /*--------------------------------.
	  | ppep_mapping_factory attributes |
	  `--------------------------------*/

	  boost::shared_ptr<node>	root_;
	};

      } // End of namespace internals.

      typedef
        tool::mapping<essid, essid, internals::ppep_mapping_factory>
        prefix_preserving_essid_policy;

    } // End of namespace wifi::frame::anon.

  } // End of namespace wifi::frame.

} // End of namespace wifi.

# include "anon_policies.hxx"

#endif // ! WIFI_FRAME_FILTER_ANON_POLICIES_HH_
