/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2008, 2009  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#ifndef WIFI_FRAME_HASH_HXX_
# define WIFI_FRAME_HASH_HXX_

# include <wipal/tool/options.hh>
# include <wipal/pcap/list_of_traces.hh>

# include <wipal/wifi/frame/unique_id/hash.hh>
# include <wipal/wifi/frame/frame.hh>
# include <wipal/wifi/frame/mgt.hh>

# include <wipal/proto/en2.hh>
# include <wipal/proto/llc.hh>
# include <wipal/proto/snap.hh>

namespace wifi
{

  namespace frame
  {

    namespace internals
    {

      hash_value::hash_value(const void* data, size_t len)
      {
	MD5(static_cast<const unsigned char*> (data), len, &this->front());
      }

      inline
      bool
      unique_udp(const udp::header* const h, const size_t caplen)
      {
	if (caplen < sizeof (*h))
	  return false;

	switch (tool::extract_big_endian_short_u(h->ports[udp::header::dp]))
	  {
	  case udp::port::OLSR:
	    /*
	     * We at least need the packet sequence number (bytes 2-3)
	     * to consider the OLSR packet to be unique.
	     */
	    return caplen >= sizeof (udp::header) + 4;

	  default:
	    return false;
	  }
      }

      inline
      bool
      unique_icmp6(const icmp6::header* const h, const size_t caplen)
      {
	if (caplen < sizeof (*h))
	  return false;

	return h->type == icmp6::type::router_adv;
      }

      inline
      bool
      unique_ip4(const ip4::header* const	h,
		 const size_t			caplen,
		 const ip4::addr* const		ignore)
      {
	if (caplen < sizeof (*h))
	  return false;

	const size_t	len = (h->headerlength & 0xF) << 2;

	if (caplen < len or (ignore and *ignore == h->addrs[ip4::header::sa]))
	  return false;

	switch (h->ipproto)
	  {
	  case ip4::proto::UDP:
	    {
	      const uint8_t* const p = reinterpret_cast<const uint8_t*> (h);

	      return unique_udp(reinterpret_cast<const udp::header*> (p + len),
				caplen - len);
	    }

	  default:
	    return false;
	  }
      }

      inline
      bool
      unique_ip6(const ip6::header* const	h,
		 const size_t			caplen,
		 const ip6::addr* const		ignore)
      {
	if (caplen < sizeof (*h) or (ignore and
				     *ignore == h->addrs[ip6::header::sa]))
	  return false;

	switch (h->nextheader)
	  {
	  case ip6::proto::ICMPv6:
	    return unique_icmp6(reinterpret_cast<const icmp6::header*> (h + 1),
				caplen - sizeof (*h));

	  case ip6::proto::UDP:
	    return unique_udp(reinterpret_cast<const udp::header*> (h + 1),
			      caplen - sizeof (*h));

	  default:
	    return false;
	  }
      }

      template <class Addr, class D>
      boost::optional<Addr>
      get(const pcapxx::frame_descriptor<D>& f)
      {
	typedef const D*					desc_ptr;
	typedef typename pcapxx::list_of_traces<D>*		list_ptr;
	typedef const typename pcapxx::list_of_traces<D>*	list_const_ptr;

	boost::optional<Addr>	r;
	const desc_ptr		d = f.desc_ptr();

	if (d)
	  {
	    const tool::opt::list&	m = d->metadata();

	    if (m.count("addr"))
	      r = Addr (m["addr"].get<std::string>());
	    else if (m.count("list_of_traces"))
	      {
		const list_const_ptr	l = m["list_of_traces"].get<list_ptr>();
		const tool::opt::list&	lm = l->metadata();

		if (lm.count("addr"))
		  r = Addr (lm["addr"].get<std::string>());
	      }
	  }

	return r;
      }

      struct hash_hooks_80211: public dissector_default_hooks
      {

	inline
	void
	probe_resp_hook(const mgt::header* frame, size_t caplen)
	{
	  if (not control_flag(3, frame))
	    result_ = hash_value (frame, caplen);
	}

	inline
	void
	beacon_hook(const mgt::header* frame, size_t caplen)
	{
	  result_ = hash_value (frame, caplen);
	}

	boost::optional<hash_value>	result_;
      };

      struct hash_hooks_80211_OLSR: public hash_hooks_80211
      {

	inline
	hash_hooks_80211_OLSR(): orig_data_ (false)
	{
	}

	inline
	void
	data_hook(const void* frame, size_t)
	{
	  if (not control_flag(3, frame))
	    orig_data_ = true;
	}

	inline
	void
	end_of_frame_headers_hook(const void*	frame,
				  size_t	caplen,
				  const void*	pload)
	{
	  typedef const llc::header*	llc_hdr;
	  typedef const snap::header*	snp_hdr;
	  typedef const ip4::header*	ip4_hdr;
	  typedef const ip6::header*	ip6_hdr;

	  if (not orig_data_)
	    return;

	  const size_t	plen = caplen - (static_cast<const uint8_t*> (pload) -
					 static_cast<const uint8_t*> (frame));
	  const size_t	hlen = sizeof (llc::header) + sizeof (snap::header);

	  if (plen < hlen)
	    return;

	  const llc::header*	l = static_cast<llc_hdr> (pload);

	  if (l->dsap != 0xAA or l->ssap != 0xAA)
	    return;

	  const snp_hdr		snap = reinterpret_cast<snp_hdr> (l + 1);
	  static const uint8_t	snap_ip4[] = { 0x00, 0x00, 0x00, 0x08, 0x00 };
	  static const uint8_t	snap_ip6[] = { 0x00, 0x00, 0x00, 0x86, 0xDD };

	  if ((not memcmp(snap, snap_ip4, sizeof (snap_ip4)) and
	       unique_ip4(reinterpret_cast<ip4_hdr> (snap + 1), plen - hlen)) or
	      (not memcmp(snap, snap_ip6, sizeof (snap_ip6)) and
	       unique_ip6(reinterpret_cast<ip6_hdr> (snap + 1), plen - hlen)))
	    result_ = hash_value (frame, caplen);
	}

      protected:
	bool	orig_data_;
      };

    } // End of namespace wifi::frame::internals.

    template <class HeaderType, class D>
    boost::optional<hsh_80211_id>
    unique_id_factory<hsh_80211_id>::
    build(const pcapxx::frame_descriptor<D>& desc, addr_mapping&)
    {
      WP_UNIQUE_ID_DISSECT(desc, d, internals::hash_hooks_80211, h);

      if (h.result_)
	return hsh_80211_id (desc.id(), h.result_.get());
      else
	return boost::none_t ();
    }

    template <class HeaderType, class D>
    boost::optional<hsh_80211_x_id>
    unique_id_factory<hsh_80211_x_id>::
    build(const pcapxx::frame_descriptor<D>& desc, addr_mapping&)
    {
      WP_UNIQUE_ID_DISSECT(desc, d, internals::hash_hooks_80211_OLSR, h);

      if (h.result_)
	return hsh_80211_x_id (desc.id(), h.result_.get());
      else
	return boost::none_t ();
    }

    template <class HeaderType, class D>
    boost::optional<hsh_en2_id>
    unique_id_factory<hsh_en2_id>::
    build(const pcapxx::frame_descriptor<D>& desc, addr_mapping&)
    {
      typedef const en2::header*	en2_hdr;
      typedef const ip4::header*	ip4_hdr;
      typedef const ip6::header*	ip6_hdr;

      const size_t	caplen = desc.pcap_header()->caplen;

      if (caplen < sizeof (en2::header))
	return boost::none_t ();

      const en2_hdr	hdr = reinterpret_cast<en2_hdr> (desc.bytes().get());
      bool		uniq;

      switch (tool::extract_big_endian_short_u(hdr->ethertype))
	{
	case en2::type::IP:
	  {
	    boost::optional<ip4::addr>	a = internals::get<ip4::addr>(desc);

	    uniq = internals::unique_ip4(reinterpret_cast<ip4_hdr> (hdr->data),
					 caplen - sizeof (*hdr),
					 a.get_ptr());
	  }
	  break;

	case en2::type::IPv6:
	  uniq = internals::unique_ip6(reinterpret_cast<ip6_hdr> (hdr->data),
				       caplen - sizeof (*hdr));
	  break;

	default:
	  uniq = false;
	  break;
	}

      if (uniq)
	return hsh_en2_id (desc.id(), internals::hash_value (hdr, caplen));
      else
	return boost::none_t ();
    }

  } // End of namespace wifi::frame.

} // End of namespace wifi.

WP_HASH_NAMESPACE_BEGIN

  inline
  size_t
  hash<wifi::frame::internals::hash_value>::
  operator () (const wifi::frame::internals::hash_value& v) const
  {
    typedef wifi::frame::internals::hash_value	hash_value;
    typedef hash_value::value_type		value_type;
    typedef hash_value::const_iterator		const_iterator;

    size_t			r = 0;
    const_iterator		i = v.begin();
    const const_iterator	e = v.end();

    while (i != e)
      for (unsigned j = 0; i != e and j < sizeof (size_t); ++i, ++j)
	r ^= (*i << (j * 8));

    return r;
  }

WP_HASH_NAMESPACE_END

#endif // ! WIFI_FRAME_HASH_HXX_
